import pytest
import itertools
from contextlib import contextmanager


@contextmanager
def _noop():
    yield

class UnitestSteps(object):
    def __init__(self, param_name=None):
        self._steps = []
        self._got_permutations = False
        self._param_name = param_name or "steps"

    def add(self, func):
        assert not self._got_permutations, "Cannot add step after got permutations"
        self._steps.append(func)

    def get_all_permutations(self, func):
        self._got_permutations = True
        return pytest.mark.parametrize(self._param_name, itertools.permutations(self._steps))(func)



import gossip
import gossip.exceptions
import pytest

from .conftest import RegisteredHook

# pylint: disable=redefined-outer-name


def test_not_now():
    with pytest.raises(gossip.NotNowException):
        gossip.not_now()


def test_wait_for():
    gossip.wait_for(True)
    gossip.wait_for(1)
    gossip.wait_for("string")

    for falsey in [False, "", None, [], ()]:
        with pytest.raises(gossip.NotNowException):
            gossip.wait_for(falsey)


def test_regular_call_order(handlers):
    _trigger(handlers)
    timestamps = [handler.last_timestamp for handler in handlers]
    assert sorted(timestamps) == timestamps


def test_dependency(handlers):
    handlers[3].depend_on(handlers[7])
    _trigger(handlers)
    timestamps = [handler.last_timestamp for handler in handlers]
    assert timestamps != sorted(timestamps)
    timestamps.append(timestamps.pop(3))
    assert timestamps == sorted(timestamps)


def test_circular_dependency(handlers):
    handlers[3].depend_on(handlers[7])
    handlers[7].depend_on(handlers[3])
    with pytest.raises(gossip.exceptions.CannotResolveDependencies):
        _trigger(handlers)
    for index, handler in enumerate(handlers):
        assert handler.called == (index not in (3, 7))


def test_reordering(handlers):  # pylint: disable=redefined-outer-name, unused-argument
    pytest.skip("n/i")


def _trigger(handlers):
    handlers[0].trigger()


@pytest.fixture
def handlers(hook_name):
    return [RegisteredHook(hook_name) for _ in range(10)]



import gossip
from gossip import Blueprint
from gossip.exceptions import UndefinedHook

import pytest


def test_multiple_functions_same_hook(blueprint, checkpoint):  # pylint: disable=unused-argument

    called = set()

    @blueprint.register('a')
    def handler1():  # pylint: disable=unused-variable
        called.add(1)

    @blueprint.register('a')
    def handler2():  # pylint: disable=unused-variable
        called.add(2)

    blueprint.install()
    gossip.trigger('a')
    assert called == set([1, 2])


def test_blueprint(blueprint, checkpoint):

    @blueprint.register('a.b')
    def handler():  # pylint: disable=unused-variable
        checkpoint()

    gossip.trigger('a.b')

    assert not checkpoint.called

    blueprint.install()

    gossip.trigger('a.b')
    assert checkpoint.called


def test_uninstall(blueprint, checkpoint):

    blueprint.register('a')(checkpoint)
    blueprint.install()
    gossip.trigger('a')
    assert checkpoint.called
    checkpoint.reset()
    blueprint.uninstall()
    gossip.trigger('a')
    assert not checkpoint.called


def test_exception_on_register(blueprint):

    gossip.get_or_create_group('group').set_strict()

    @blueprint.register('group.a')
    def func():  # pylint: disable=unused-variable
        pass

    @blueprint.register('x')
    def func():  # pylint: disable=function-redefined
        pass

    with pytest.raises(UndefinedHook):
        blueprint.install()

    assert len(gossip.get_all_registrations()) == 0


def test_blueprint_register_prefix(blueprint, checkpoint):

    @gossip.register('a')
    def handler():  # pylint: disable=unused-variable
        pass

    @blueprint.register('a')
    def handler():  # pylint: disable=function-redefined
        checkpoint()

    assert len(gossip.get_all_registrations()) == 1

    blueprint.install(group='group')

    gossip.trigger('a')
    assert not checkpoint.called

    gossip.trigger('group.a')

    assert checkpoint.called



@pytest.fixture
def blueprint():
    return Blueprint()



# pylint: disable=unused-argument
from munch import Munch

import gossip
import gossip.groups
import pytest
from gossip.exception_policy import Inherit

from .conftest import RegisteredHook


def test_exception_from_internal_hook(registered_hook):
    registered_hook.fail_when_called()
    called = Munch(count=0)

    # pylint: disable=unused-variable
    @gossip.register("gossip.on_handler_exception")
    def handle_exception(handler, exception, hook):
        called.count += 1
        raise CustomException()

    with pytest.raises(CustomException):
        registered_hook.trigger()

    assert called.count == 1


class CustomException(Exception):
    pass


def test_gossip_exception_caught_hook(registered_hook):
    registered_hook.fail_when_called()
    error = Munch(caught=False)

    @gossip.register("gossip.on_handler_exception")
    def handle_exception(handler, exception, hook):  # pylint: disable=unused-variable, unused-argument
        assert handler is registered_hook.func
        assert exception[0] is registered_hook.exception_class
        assert isinstance(exception[1], registered_hook.exception_class)
        error.caught = True

    with pytest.raises(registered_hook.exception_class):
        registered_hook.trigger()

    assert error.caught


@pytest.mark.parametrize("invalid_policy", [object(), 1, None, gossip.RaiseImmediately])
def test_cannot_set_invalid_policy(invalid_policy):
    with pytest.raises(ValueError):
        gossip.set_exception_policy(invalid_policy)


def test_global_cannot_be_inherit():
    with pytest.raises(RuntimeError):
        gossip.set_exception_policy(Inherit())


def test_default_global_group_policy():
    assert isinstance(gossip.get_global_group()
                      .get_exception_policy(), gossip.RaiseImmediately)


def test_inherit_policy(exception_handling_policy):
    group = gossip.groups.Group("some_group", parent=gossip.get_global_group())
    gossip.set_exception_policy(exception_handling_policy)
    group.set_exception_policy(Inherit())
    assert group.get_exception_policy() is exception_handling_policy


def test_inherit_policy_changing_parent():
    gossip.define("group.hook")
    group = gossip.get_group("group")
    for policy in (gossip.RaiseImmediately(), gossip.RaiseDefer(), gossip.IgnoreExceptions()):
        gossip.set_exception_policy(policy)
        assert group.get_exception_policy() is policy


def test_group_returns_to_default_after_reset():
    gossip.get_global_group().set_exception_policy(gossip.IgnoreExceptions())
    gossip.get_global_group().reset()
    test_default_global_group_policy()


def test_get_exception_policy(exception_handling_policy):
    group = gossip.groups.Group("some_group", parent=gossip.get_global_group())
    group.set_exception_policy(exception_handling_policy)
    assert group.get_exception_policy() is exception_handling_policy


def test_exception_handling(hook, error_prone_handlers, exception_handling_policy):
    hook.group.set_exception_policy(exception_handling_policy)
    if isinstance(exception_handling_policy, gossip.IgnoreExceptions):
        hook(**error_prone_handlers[0].kwargs)
    else:
        with pytest.raises(error_prone_handlers[0].exception_class):
            hook(**error_prone_handlers[0].kwargs)

    assert error_prone_handlers[0].called
    for handler in error_prone_handlers[1:]:
        assert handler.called == (
            not isinstance(exception_handling_policy, gossip.RaiseImmediately))


@pytest.fixture(
    params=[gossip.RaiseDefer, gossip.RaiseImmediately, gossip.IgnoreExceptions])
def exception_handling_policy(request):
    return request.param()


@pytest.fixture
def error_prone_handlers(hook_name):
    returned = []
    num_handlers = 10
    for handler_index in range(num_handlers):
        handler = RegisteredHook(hook_name)
        if handler_index in (0, 5, num_handlers - 1):
            handler.fail_when_called()
        returned.append(handler)

    return returned



from contextlib import contextmanager

import warnings
import gossip


def test_deprecated_hook():
    with python_warnings_recording() as recorded:

        hook = gossip.define('hook', deprecated=True)
        assert recorded == []

        @hook.register
        def handler():  # pylint: disable=unused-variable
            pass

        [rec] = recorded

        assert rec.filename == __file__


@contextmanager
def python_warnings_recording():
    warnings.simplefilter('always')
    with warnings.catch_warnings(record=True) as recorded:
        yield recorded



import gossip


def test_normal_hooks_can_reenter(counter):
    counter.set(5)

    @gossip.register('reentrant_hook')
    def handler(): # pylint: disable=unused-variable
        counter.sub(1)
        if counter:
            gossip.trigger('reentrant_hook')

    gossip.trigger('reentrant_hook')
    assert counter == 0


def test_non_reentrant_hooks(counter):
    hook_name = 'nonreentrant_hook'


    @gossip.register(hook_name, reentrant=False)
    def handler(): # pylint: disable=unused-variable
        counter.add(1)
        gossip.trigger(hook_name)

    gossip.trigger(hook_name)
    assert counter == 1



from __future__ import print_function
import os
import doctest

import pytest

PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

@pytest.mark.parametrize("path", [
    os.path.join(path, filename)
    for path, _, filenames in os.walk(os.path.join(PROJECT_ROOT, 'doc'))
    for filename in filenames
    if filename.endswith((".rst", ".md"))])
def test_doctests(path):
    assert os.path.exists(path)
    result = doctest.testfile(path, module_relative=False)
    assert result.failed == 0



import gossip
import gossip.groups
import pytest
from gossip.exceptions import NameAlreadyUsed, GroupNotFound

# pylint: disable=unused-variable

def test_trigger_no_hook_registered():
    result = gossip.trigger("unregistered")  # pylint: disable=assignment-from-no-return
    assert result is None


def test_create_group_twice():
    group_name = "group_name"
    gossip.create_group(group_name)

    with pytest.raises(NameAlreadyUsed):
        gossip.create_group(group_name)


def test_get_or_create_group():
    group_name = "group_name"
    with pytest.raises(GroupNotFound):
        gossip.get_group(group_name)

    group = gossip.get_or_create_group(group_name)
    assert isinstance(group, gossip.groups.Group)
    assert group is gossip.get_group(group_name)


def test_hook_documentation():
    docstring = "fdkjfkdjfd"
    gossip.define("some_hook", doc=docstring)
    assert gossip.get_hook("some_hook").doc == docstring


def test_register_trigger(registered_hook):
    assert registered_hook.works()


def test_define_twice(hook_name):
    gossip.define(hook_name)
    with pytest.raises(NameAlreadyUsed):
        gossip.define(hook_name)


def test_undefine(hook_name):
    gossip.define(hook_name)

    @gossip.register(hook_name)
    def handler():
        pass

    gossip.get_hook(hook_name).undefine()
    gossip.define(hook_name)

    assert not gossip.get_hook(hook_name).get_registrations()


def test_get_hook(hook_name):
    with pytest.raises(LookupError):
        gossip.get_hook(hook_name)
    hook = gossip.define(hook_name)
    assert gossip.get_hook(hook_name) is hook


def test_register_function_already_has_gossip_attr(hook_name):
    def func():
        pass

    func.gossip = old_obj = object()

    gossip.register(func=func, hook_name=hook_name)
    assert func.gossip is old_obj


def test_unregister(registered_hook):
    assert registered_hook.works()
    registered_hook.func.gossip.unregister()
    assert not registered_hook.works()

def test_unregister_all_on_hook(registered_hooks):
    assert all(r.works() for r in registered_hooks)
    gossip.get_hook(registered_hooks[0].name).unregister_all()
    assert not registered_hooks[0].works()
    assert all(r.works() for r in registered_hooks[1:])


def test_global_group_reset_delets_all_group(registered_hooks):
    assert gossip.get_groups()
    gossip.get_global_group().reset()
    assert not gossip.get_groups()
    assert all(not r.works() for r in registered_hooks)
    assert not gossip.get_global_group().get_subgroups()


def test_global_group_is_same(registered_hooks):  # pylint: disable=unused-argument
    global_group = gossip.get_global_group()
    global_group.reset()
    assert gossip.get_global_group() is global_group


def test_hook_name_taken_by_group():
    @gossip.register("group_name.hook")
    def func():
        pass

    with pytest.raises(NameAlreadyUsed):
        @gossip.register("group_name")
        def func2():
            pass

    assert "group_name" not in gossip.registry.hooks


def test_group_name_taken_by_hook():
    @gossip.register("group_name")
    def func():
        pass

    with pytest.raises(NameAlreadyUsed):
        @gossip.register("group_name.hook")
        def func2():
            pass

    assert "group_name" not in gossip.registry.groups


def test_register_class_and_instance_methods():

    class MyClass(object):

        @gossip.register("hook")
        @classmethod
        def class_method(cls):
            pass

        @gossip.register("hook")
        @staticmethod
        def static_method():
            pass

        def regular_method(self):
            pass

    m = MyClass()

    gossip.register("hook")(m.regular_method)


def test_nested_groups():

    @gossip.register("a.b.c")
    def handler():
        pass

    assert gossip.get_group("a").name == "a"
    assert gossip.get_group("a.b").name == "b"


def test_mute_trigger(checkpoint):

    @gossip.register('a.b.c')
    def handler():
        checkpoint()

    with gossip.mute_context(['a.b.c']):
        gossip.trigger('a.b.c')

    assert not checkpoint.called

    gossip.trigger('a.b.c')

    assert checkpoint.called


@pytest.mark.parametrize('arg', ['', 'name', 2, 2.0, True])
def test_mute_accepts_only_lists(arg):

    with pytest.raises(TypeError):
        with gossip.mute_context(arg):
            pass



import itertools

from .exceptions import CannotResolveDependencies
from .helpers import DONT_CARE, FIRST

def topological_sort_registrations(registrations, unconstrained_priority=DONT_CARE):
    graph = _build_dependency_graph(registrations, unconstrained_priority=unconstrained_priority)
    returned_indices = _topological_sort(range(len(registrations)), graph)
    assert len(returned_indices) == len(registrations)
    return [registrations[idx] for idx in returned_indices]


def _topological_sort(indices, graph):
    independent = sorted(set(indices) - set(m for n, m in graph), reverse=True)
    returned = []
    while independent:
        n = independent.pop()
        returned.append(n)
        for m in indices:
            edge = (n, m)
            if m == n:
                assert edge not in graph
                continue
            if edge in graph:
                graph.remove(edge)
                # check if m is now independent
                for edge in graph:
                    if edge[1] == m:
                        # not indepdendent
                        break
                else:
                    # no other incoming edges to m
                    independent.append(m)
    if graph:
        raise CannotResolveDependencies('Cyclic dependency detected')
    return returned


def _build_dependency_graph(registrations, unconstrained_priority):
    providers_by_name = {}
    for index, registration in enumerate(registrations):
        for name in registration.provides:
            providers = providers_by_name.get(name)
            if providers is None:
                providers = providers_by_name[name] = []
            providers.append(index)

    graph = set()
    for needer_index, registration in enumerate(registrations):
        for need in registration.needs:
            for provider_index in providers_by_name.get(need, []):
                graph.add((provider_index, needer_index))


    if unconstrained_priority != DONT_CARE:
        caring_indices = set([idx for idx, r in enumerate(registrations) if r.needs or r.provides])
        non_caring_indices = set(range(len(registrations))) - caring_indices
        for caring_index, uncaring_index in itertools.product(caring_indices, non_caring_indices):
            if unconstrained_priority == FIRST:
                pair = (uncaring_index, caring_index)
            else:
                pair = (caring_index, uncaring_index)
            graph.add(pair)


    return graph



import pytest

from gossip import Hook, get_global_group, define, trigger, register


def test_pre_trigger_callback(checkpoint):
    called_args = []

    hook = define('somehook')

    @hook.add_pre_trigger_callback
    def pre_trigger_callback(registration, kwargs):  # pylint: disable=unused-variable
        called_args.append((registration, kwargs))

    assert not called_args

    trigger('somehook')

    assert not called_args

    @register('somehook')
    def callback(**kwargs):  # pylint: disable=unused-argument
        checkpoint()

    assert not called_args

    trigger('somehook', x=1, y=2, z=3)

    assert len(called_args) == 1
    assert called_args[0][-1] == {'x': 1, 'y': 2, 'z': 3}
    assert called_args[0][0].func is callback


def test_remove_pre_trigger_callback():
    # pylint: disable=protected-access
    hook = define('hook')

    @hook.add_pre_trigger_callback
    def callback():
        pass

    assert len(hook._pre_trigger_callbacks) == 1
    hook.remove_pre_trigger_callback(callback)
    assert not hook._pre_trigger_callbacks


@pytest.mark.parametrize('stringify', [str, repr])
def test_hook_str_repr(stringify):
    assert stringify(Hook(get_global_group(), "some_name", arg_names=("a", "b"))) == \
        "<Hook some_name(a, b)>"



from munch import Munch

import gossip
import pytest


@pytest.fixture
def registrations():
    @gossip.register("group.hook1")
    def handler1():
        pass

    @gossip.register("group.hook2", token="token1")
    def handler2():
        pass

    @gossip.register("group2.subgroup.hook3", token="token1")
    def handler3():
        pass

    @gossip.register("group3.subgroup.subgroup2.hook4", token="token2")
    def handler4():
        pass

    return Munch(handler1=handler1, handler2=handler2, handler3=handler3, handler4=handler4)


def test_token_registration_unregistration(registrations):

    assert len(_get_all_registrations()) == 4
    gossip.unregister_token("token2")
    assert len(_get_all_registrations()) == 3
    assert not registrations.handler4.gossip.is_active()


def test_group_only_token_unregistration(registrations):  # pylint: disable=unused-argument

    gossip.get_group("group2").unregister_token("token1")
    assert len(_get_all_registrations()) == 3


def test_token_unregister_from_hook(checkpoint, token, hook_name):
    # pylint: disable=unused-variable

    @gossip.register(hook_name, token=token)
    def handler():
        gossip.unregister_token(token)
    gossip.register(hook_name)(checkpoint)

    gossip.trigger(hook_name)
    assert checkpoint.called


def test_token_unregister_from_hook_multiple_token_registrants(checkpoint, token, hook_name):
    # pylint: disable=unused-variable

    @gossip.register(hook_name, token=token)
    def handler():
        gossip.unregister_token(token)
    gossip.register(hook_name)(checkpoint)
    # the second token registration should not be called!
    gossip.register(hook_name, token=token)(checkpoint)

    gossip.trigger(hook_name)
    assert checkpoint.num_times == 1



def _get_all_registrations():
    return [registration for hook in gossip.get_all_hooks() for registration in hook.get_registrations()]



import functools
import logbook
import sys
from collections import OrderedDict
from contextlib import contextmanager
from types import GeneratorType
from sentinels import Sentinel

from . import groups, registry
from ._compat import itervalues, string_types
from .exceptions import (CannotResolveDependencies, HookNotFound,
                         NameAlreadyUsed, NotNowException, UndefinedHook,
                         UnsupportedHookTags)
from .registration import Registration
from .utils import topological_sort_registrations

from vintage import warn_deprecation

_logger = logbook.Logger(__name__)

_REGISTER_NO_OP = Sentinel('REGISTER_NO_OP')


class Hook(object):

    def __init__(self, group, name, arg_names=None, doc=None, deprecated=False):
        super(Hook, self).__init__()
        self.group = group
        self.name = name
        self.tags = None
        if self.group.is_global():
            self.full_name = name
        else:
            self.full_name = "{0}.{1}".format(self.group.full_name, self.name)
        registry.hooks[self.full_name] = self
        self._registrations = []
        self._empty_regisrations = []
        self._arguments = self._normalize_arguments(arg_names)
        self._trigger_internal_hooks = self.full_name != "gossip.on_handler_exception"
        self._defined = False
        self._pre_trigger_callbacks = []
        self._unmet_deps = frozenset()
        self.doc = doc
        self.deprecated = deprecated

    def _get_registration_list_from_func(self, func):
        if func is _REGISTER_NO_OP:
            return self._empty_regisrations
        return self._registrations

    def _normalize_arguments(self, arg_names):
        if arg_names is None:
            return None

        if isinstance(arg_names, dict):
            return OrderedDict(arg_names)

        return OrderedDict((arg_name, None) for arg_name in arg_names)

    def add_pre_trigger_callback(self, callback):
        self._pre_trigger_callbacks.append(callback)
        return callback

    def remove_pre_trigger_callback(self, callback):
        self._pre_trigger_callbacks.remove(callback)

    def undefine(self):
        self.group.remove_child(self.name)
        registry.hooks.pop(self.full_name)

    def set_tags(self, tags):
        assert not self.tags, "Cannot override exists tags {0} with {1}".format(
            self.tags, tags)
        self.tags = tags

    def get_registrations(self):
        return list(self._registrations)

    def mark_defined(self):
        self._defined = True

    def is_defined(self):
        return self._defined

    def get_argument_names(self):
        return tuple(self._arguments or ())

    def __call__(self, **kwargs):
        return self.trigger(kwargs)

    def validate_tags(self, tags, is_strict=None):
        if is_strict is None:
            is_strict = self.group.is_strict()
        if not is_strict:
            return
        normalize_tags = lambda tag_: set(tag_) if tag_ is not None else set()
        extra_tags = normalize_tags(tags) - normalize_tags(self.tags)
        if extra_tags:
            raise UnsupportedHookTags("hook {0} support {1} tags, not: {2}".format(
                self.full_name, self.tags, extra_tags))

    def validate_kwargs(self, kwargs):
        if self._arguments is None or not self.group.is_strict():
            return

        unknown = set(kwargs) - set(self._arguments)
        if unknown:
            raise TypeError('Unknown arguments specified: {}'.format(', '.join(unknown)))

        for arg_name, arg_types in self._arguments.items():
            if arg_name not in kwargs:
                raise TypeError('Missing argument {!r}'.format(arg_name))
            if arg_types is not None and not isinstance(kwargs[arg_name], arg_types):
                raise TypeError('Incorrect type for argument {}. Expected {!r}, got {!r}'.format(
                    arg_name, arg_types, type(kwargs[arg_name])))


    def validate_strict(self, registrations_to_validate=None):
        if not self._defined:
            raise UndefinedHook(
                "hook {0} wasn't defined yet".format(self.full_name))
        if registrations_to_validate is None:
            registrations_to_validate = self._registrations + self._empty_regisrations
        for registration in registrations_to_validate:
            self.validate_tags(registration.tags, is_strict=True)

    def register(self, func, token=None, tags=None, needs=None, provides=None, **kwargs):
        """Registers a new handler to this hook
        """
        if self.deprecated:
            warn_deprecation('Hook {0} is deprecated!'.format(self.full_name), frame_correction=+1)
        new_registration = Registration(func, self, token=token, tags=tags, needs=needs, provides=provides, **kwargs)
        if self.group.is_strict():
            self.validate_strict([new_registration])
        need_sorting = self._registrations and new_registration.priority > self._registrations[-1].priority
        self._get_registration_list_from_func(func).append(new_registration)
        if need_sorting:
            self._registrations.sort(key=Registration.get_priority, reverse=True) # sort is stable, so order among registration isn't disturbed
        if new_registration.needs or new_registration.provides:
            try:
                self.recompute_call_order()
            except:
                self._registrations.pop()
                raise
        return new_registration

    def register_no_op(self, **kwargs):
        if kwargs.get('needs'):
            raise NotImplementedError("Cannot define 'needs' for register_no_op")
        return self.register(func=_REGISTER_NO_OP, **kwargs)

    def recompute_call_order(self):
        self._registrations = topological_sort_registrations(self._registrations, unconstrained_priority=self.group.get_unconstrained_handler_priority())
        self._unmet_deps = frozenset(n for r in self._registrations for n in r.needs) - \
                           frozenset(p for r in (self._registrations + self._empty_regisrations) for p in r.provides)


    def unregister(self, registration):
        assert registration.hook is self
        self._get_registration_list_from_func(registration.func).remove(registration)
        registration.hook = None
        self.recompute_call_order()

    def unregister_all(self):
        del self._registrations[:]
        del self._empty_regisrations[:]
        self.recompute_call_order()

    def trigger(self, kwargs, tags=None):
        if self._unmet_deps:
            deps_str = ', '.join([str(dep) for dep in self._unmet_deps])
            raise CannotResolveDependencies('Hook {0!r} has unmet dependencies: {1}'.format(self, deps_str))
        if self.full_name in _muted_stack[-1]:
            _logger.debug("Hook {0!r} muted, skipping trigger", self)
            return

        self.validate_tags(tags)
        self.validate_kwargs(kwargs)
        exception_policy = self.group.get_exception_policy()

        registrations = self._registrations
        deferred = []

        with exception_policy.context() as ctx:
            while True:
                any_resolved = False
                for registration in list(registrations):
                    if not registration.valid:
                        continue
                    if not registration.has_tags(tags):
                        continue
                    try:
                        exc_info = self._call_registration(
                            registration, kwargs)
                    except NotNowException:
                        deferred.append(registration)
                        continue
                    else:
                        any_resolved = True
                    if exc_info is not None:
                        exception_policy.handle_exception(ctx, exc_info)
                if deferred:
                    if not any_resolved:
                        raise CannotResolveDependencies(
                            "Cannot resolve handler dependencies for {0}".format(self))
                    registrations = deferred
                    deferred = []
                else:
                    break

    def _call_registration(self, registration, kwargs):
        if registration.is_being_called() and not registration.reentrant:
            return
        exc_info = None
        for callback in self._pre_trigger_callbacks:
            callback(registration, kwargs)
        try:
            registration(**kwargs)
        except NotNowException:
            raise
        except Exception:  # pylint: disable=broad-except
            exc_info = sys.exc_info()
            if self._trigger_internal_hooks:
                trigger("gossip.on_handler_exception",
                        handler=registration.func, exception=exc_info, hook=self)
            _logger.debug("Exception occurred while calling {0}", registration, exc_info=exc_info)
        return exc_info

    def __repr__(self):
        return "<Hook {0}({1})>".format(self.name, ", ".join(self._arguments or ()))


def trigger(hook_name, **kwargs):
    """Triggers a hook by name, causing all of its handlers to be called
    """
    trigger_with_tags(hook_name, kwargs, None)


def trigger_with_tags(hook_name, kwargs=None, tags=None):
    """Triggers a hook by name, specifying the tags to use when triggering.
    If the hook receives keyword arguments, they have to be passed using the *kwargs* parameter, and not via
    direct keyword arguments
    """
    hook = registry.hooks.get(hook_name)
    if hook is not None:
        hook.trigger(kwargs or {}, tags)


def get_or_create_hook(hook_name, **kwargs):
    try:
        return get_hook(hook_name)
    except HookNotFound:
        return create_hook(hook_name, **kwargs)


def get_hook(hook_name):
    """Gets a hook by its name

    :raises: :class:`gossip.exceptions.HookNotFound` if the hook wasn't defined already
    """
    try:
        return registry.hooks[hook_name]
    except KeyError:
        raise HookNotFound("Hook {0} does not exist".format(hook_name))


def create_hook(hook_name, **kwargs):
    """Creates a hook with the given full name, creating intermediate groups if necessary
    """
    if hook_name in registry.hooks:
        raise NameAlreadyUsed(
            "A hook named {0} already exists. Cannot create a hook with the same name".format(hook_name))

    if hook_name in registry.groups:
        raise NameAlreadyUsed(
            "A group named {0} already exists. Cannot create a hook with the same name".format(hook_name))

    if "." in hook_name:
        group_name, hook_base_name = hook_name.rsplit(".", 1)
    else:
        group_name = None
        hook_base_name = hook_name

    group = groups.get_or_create_group(group_name)
    hook = Hook(group, hook_base_name, **kwargs)
    group.add_child(hook_base_name, hook)
    return hook


def define(hook_name, **kwargs):
    """Defines a new hook with the given name

    :returns: The :class:`gossip.hook.Hook` object created
    """
    tags = kwargs.pop('tags', None)
    returned = get_or_create_hook(hook_name, **kwargs)
    if returned.is_defined():
        raise NameAlreadyUsed("Hook {0} is already defined".format(hook_name))
    if tags:
        returned.set_tags(tags)
    returned.mark_defined()
    return returned


def register(func=None, hook_name=None, token=None, tags=None, needs=None, provides=None, reentrant=True, toggles_on=None, toggles_off=None, priority=0):
    """Registers a new function to a hook

    :param hook_name: full name of hook to register to
    :param token: token to register with. This can be used to later unregister a group of handlers which have a
           specific token
    :param needs: list of keywords (strings) that this registration needs, causing any hook that provides any of them to happen before this registration
    :param provides: list of keywords (strings) that this registration provides
    :param reentrent: specifies whether this hook can reenter (i.e. be called in recursion)
    :param toggles_on: specifies a toggle object to turn on when calling this registration. The registration will not be called if the toggle isn't off
    :param toggles_off: specifies a toggle object to turn off when calling this registration. The registration will not be called if the toggle isn't on
    :returns: The function (for decorator chaining)

    """
    if isinstance(func, string_types):
        assert hook_name is None
        hook_name = func
        func = None

    if func is None:
        return functools.partial(
            register,
            hook_name=hook_name,
            token=token,
            tags=tags,
            needs=needs,
            provides=provides,
            reentrant=reentrant,
            toggles_on=toggles_on,
            toggles_off=toggles_off,
            priority=priority,
        )
    assert hook_name is not None
    registration = get_or_create_hook(
        hook_name).register(
            func, token=token, tags=tags, needs=needs, provides=provides, reentrant=reentrant,
            toggles_on=toggles_on, toggles_off=toggles_off,
            priority=priority,
        )
    assert registration
    return func


def unregister_all(hook_name):
    """
    Unregisters all handlers from the given hook name
    """
    try:
        hook = get_hook(hook_name)
    except HookNotFound:
        return
    hook.unregister_all()


def get_all_hooks():
    return list(itervalues(registry.hooks))

def get_all_registrations():
    return [reg for hook in get_all_hooks()
            for reg in hook.get_registrations()]


_muted_stack = [set()]


@contextmanager
def mute_context(hook_names):
    """A context manager, during the execution of which the specified hook names will be ignored. Any code that tries to
    trigger these hooks will trigger no callback.

    :type hook_names: list, tuple or generator of full hook names to mute
    """
    if not isinstance(hook_names, (list, tuple, GeneratorType)):
        raise TypeError('hook names to mute must be a list or a tuple')
    _muted_stack.append(_muted_stack[-1] | set(hook_names))
    try:
        yield
    finally:
        _muted_stack.pop(-1)



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "gossip", "__version__.py")) as version_file:
    exec(version_file.read())  # pylint: disable=W0122

_INSTALL_REQUIERS = [
    'logbook>=0.12.0',
    'vintage',
    'sentinels',
]

setup(name="gossip",
      classifiers=[
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
      ],
      description="Signaling and hooking library",
      license="BSD3",
      author="Rotem Yaari",
      author_email="vmalloc@gmail.com",
      version=__version__,  # pylint: disable=E0602
      packages=find_packages(exclude=["tests"]),

      url="https://github.com/vmalloc/gossip",

      install_requires=_INSTALL_REQUIERS,
      scripts=[],
      namespace_packages=[]
     )



from munch import Munch

import gossip
from gossip.exceptions import UndefinedHook
from .utils import UnitestSteps, _noop
import pytest


def test_undefined_hooks_have_no_side_effects():
    assert gossip.get_all_registrations() == []
    gossip.get_or_create_group('group').set_strict()

    for _ in range(5): # pylint: disable=redefined-builtin
        with pytest.raises(UndefinedHook):
            @gossip.register('group.hook')
            def handler():  # pylint: disable=unused-variable
                pass
    assert gossip.get_all_registrations() == []


steps = UnitestSteps()

@steps.add
def _make_group_strict(ctx):
    if ctx.registered_undefined_hook or (ctx.registered_defined_hook and not ctx.defined_hook):
        raises_context = pytest.raises(UndefinedHook)
    else:
        raises_context = _noop()

    with raises_context:
        gossip.get_or_create_group("group").set_strict()
        ctx.made_strict = True

@steps.add
def _register_undefined_hook(ctx):
    if ctx.made_strict:
        raises_context = pytest.raises(UndefinedHook)
    else:
        raises_context = _noop()
    with raises_context:
        @gossip.register('group.undefined_hook')
        def handler():  # pylint: disable=unused-variable
            pass
    ctx.registered_undefined_hook = True

@steps.add
def _register_defined_hook(ctx):
    if ctx.made_strict and not ctx.defined_hook:
        raises_context = pytest.raises(UndefinedHook)
    else:
        raises_context = _noop()
    with raises_context:
        @gossip.register('group.defined_hook')
        def handler():  # pylint: disable=unused-variable
            pass
    ctx.registered_defined_hook = True

@steps.add
def _define_hook(ctx):
    gossip.define("group.defined_hook")
    ctx.defined_hook = True


@steps.get_all_permutations
def test_strict_groups(steps):
    ctx = Munch(
        made_strict=False,
        defined_hook=False,
        registered_undefined_hook=False,
        registered_defined_hook=False,
    )
    for step_func in steps:
        step_func(ctx)

def test_make_group_not_strict():
    group = gossip.get_or_create_group('group')
    subgroup = gossip.get_or_create_group('group.sub')
    assert not group.is_strict()
    assert not subgroup.is_strict()

    group.set_strict()
    assert group.is_strict()
    assert subgroup.is_strict()

    group.set_strict(False)
    assert not group.is_strict()
    assert not subgroup.is_strict()



FIRST = 'FIRST'
LAST = 'LAST'
DONT_CARE = 'DONT_CARE'


class Toggle(object):

    def __init__(self):
        super(Toggle, self).__init__()
        self._on = False

    def is_on(self):
        return self._on

    def is_off(self):
        return not self._on

    def toggle(self):
        self._on = not self._on



import gossip

import pytest


def test_non_strict_hook_doesnt_validate(request):
    hook = gossip.define('somegroup.somehook', arg_names={'x': None, 'y': int})
    request.addfinalizer(hook.undefine)

    hook.trigger(kwargs={})

def test_strict_hook_without_argument_defs_doesnt_validate(request):
    group = gossip.get_or_create_group('somegroup')
    group.set_strict()
    request.addfinalizer(group.undefine)
    hook = gossip.define('somegroup.somehook')
    request.addfinalizer(hook.undefine)

    hook.trigger(kwargs={'bla': 2, 'some': 'string', 'and_an': object()})



@pytest.mark.parametrize('kwargs', [
    {},
    {'x': 1},
    {'x': 1, 'y': 2},
])
def test_strict_hook_validate_arg_names(validating_hook, kwargs):
    with pytest.raises(TypeError) as excinfo:
        validating_hook.trigger(kwargs=kwargs)

    excinfo.match('Missing argument .*')


@pytest.mark.parametrize('kwargs', [
    {'x': 1, 'y': 'str', 'z': 's'},
    {'x': 1, 'y': 'str', 'z': 's'},
    {'x': 1, 'y': 2, 'z': 3},
    {'x': 1, 'y': 's', 'z': 2.0},
])
def test_strict_hook_validate_arg_types(validating_hook, kwargs):
    with pytest.raises(TypeError) as excinfo:
        validating_hook.trigger(kwargs=kwargs)

    excinfo.match('Incorrect type')


@pytest.mark.parametrize('kwargs', [
    {'x': 1, 'y': 2, 'z': 's'},
    {'x': 1, 'y': 3, 'z': 3.0},
    {'x': object(), 'y': 2, 'z': 3.2},
])
def test_strict_hook_correct_arg_names_and_types(validating_hook, kwargs):
    validating_hook.trigger(kwargs=kwargs)



@pytest.fixture
def validating_hook(request):

    group = gossip.get_or_create_group('some_strict_group')
    group.set_strict()
    request.addfinalizer(group.undefine)

    hook = gossip.define('some_strict_group.some_hook', arg_names={'x': None, 'y': int, 'z': (str, float)})
    request.addfinalizer(hook.undefine)
    return hook



import pytest

import gossip
from gossip.exceptions import CannotResolveDependencies


def test_needs_provides_simple(timeline):
    evt1 = timeline.register(needs=['something'])
    evt2 = timeline.register(provides=['something'])
    timeline.trigger()
    assert evt1.timestamp > evt2.timestamp


def test_unmet_deps_after_unregister(hook, counter):
    reg = hook.register(counter.plus_one, provides=['something'])
    hook.register(counter.plus_one, needs=['something'])

    reg.unregister()
    with pytest.raises(CannotResolveDependencies):
        hook.trigger({})
    assert counter.get() == 0


def test_unmet_deps_resolved_after_unregister_all(hook, counter):
    hook.register(counter.plus_one, needs=['1'])
    hook.register(counter.plus_one, needs=['2'])
    hook.register_no_op(provides=['3'])
    assert len(hook._registrations) == 2  # pylint: disable=protected-access
    assert len(hook._empty_regisrations) == 1  # pylint: disable=protected-access
    assert len(hook._unmet_deps) == 2  # pylint: disable=protected-access

    hook.unregister_all()
    assert len(hook._registrations) == 0  # pylint: disable=protected-access
    assert len(hook._empty_regisrations) == 0  # pylint: disable=protected-access
    assert len(hook._unmet_deps) == 0  # pylint: disable=protected-access

    hook.trigger({})
    assert counter.get() == 0


def test_needs_provides_complex(timeline):
    events = [
        timeline.register(provides=['0'], needs=['1']),
        timeline.register(provides=['1'], needs=['2', '3']),
        timeline.register(provides=['2'], needs=['5']),
        timeline.register(provides=['3'], needs=['5']),
        timeline.register(provides=['4'], needs=['0']),
        timeline.register(provides=['5']),
    ]
    timeline.trigger()

    assert sorted(events, key=lambda e: e.timestamp) == [
        events[5], events[3], events[2], events[1], events[0], events[4]
    ]


def test_needs_provides_cyclic_simple(timeline):
    timeline.register(needs=['1'], provides=['0'])
    with pytest.raises(CannotResolveDependencies):
        timeline.register(needs=['0'], provides=['1'])
    # make sure the cyclic registration was not added
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 1  # pylint: disable=protected-access


def test_needs_provides_cyclic_complex(timeline):
    timeline.register(needs=['1'], provides=['0'])
    timeline.register(needs=['2'], provides=['1'])
    timeline.register(needs=['3'], provides=['2'])
    timeline.register(needs=['4'], provides=['3'])
    with pytest.raises(CannotResolveDependencies):
        timeline.register(needs=['0'], provides=['4'])
    # make sure the cyclic registration was not added
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 4  # pylint: disable=protected-access


def test_register_no_op_does_not_allow_needs_parameter(hook):
    with pytest.raises(NotImplementedError):
        hook.register_no_op(needs=['something'])
    assert len(hook._unmet_deps) == 0  # pylint: disable=protected-access


def test_register_no_op_solves_dependencies(timeline):
    timeline.register(needs=['1'], provides=['0'])
    timeline.register(needs=['2'], provides=['1'])
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 2  # pylint: disable=protected-access

    with pytest.raises(CannotResolveDependencies):
        timeline.trigger()

    timeline.get_hook().register_no_op(provides=['2'])
    # Make sure "register_no_op" is not registred as "regular" callback
    assert len(gossip.get_hook(timeline.hook_name)._registrations) == 2  # pylint: disable=protected-access
    assert len(gossip.get_hook(timeline.hook_name)._empty_regisrations) == 1  # pylint: disable=protected-access

    timeline.trigger()


def test_unregister_no_op_remove_dependencies(timeline):
    timeline.register(needs=['1'], provides=['0'])
    timeline.register(needs=['2'], provides=['1'])

    hook = timeline.get_hook()
    registration = hook.register_no_op(provides=['2'])
    timeline.trigger()

    registration.unregister()
    with pytest.raises(CannotResolveDependencies):
        gossip.trigger(hook.full_name)


@pytest.mark.parametrize('priority', [gossip.FIRST, gossip.DONT_CARE, gossip.LAST])
@pytest.mark.parametrize('use_parent_group', [True, False])
def test_needs_provides_unconstrained_priority(timeline, priority, use_parent_group):
    dontcare = timeline.register()
    evt1 = timeline.register(needs=['a'], provides=['b'])
    evt2 = timeline.register(provides=['a'])
    if use_parent_group:
        group = timeline.get_parent_group()
    else:
        group = timeline.get_group()

    group.set_unconstrained_handler_priority(priority)
    timeline.trigger()

    if priority in (gossip.FIRST, gossip.DONT_CARE):
        assert dontcare.timestamp < evt2.timestamp < evt1.timestamp
    else:
        assert evt2.timestamp < evt1.timestamp < dontcare.timestamp


def test_policy_gets_reset():
    gossip.get_global_group().set_unconstrained_handler_priority(gossip.FIRST)
    gossip.get_global_group().reset()
    assert gossip.get_global_group().get_unconstrained_handler_priority() == gossip.DONT_CARE


def test_unmet_dependencies(timeline):
    # pylint: disable=unused-variable
    evt = timeline.register(needs=['a'])
    with pytest.raises(CannotResolveDependencies):
        timeline.trigger()

    evt2 = timeline.register(provides=['a'])
    timeline.trigger() # ok



import itertools

_call_timestamps = itertools.count()


def test_regular_ordering(hook):

    handler1, handler2, handler3 = handlers = [create_handler() for _ in range(3)] # pylint: disable=redefined-builtin

    for h in handlers:
        hook.register(h)

    hook.trigger(kwargs={})

    assert handler1.call_timestamp < handler2.call_timestamp < handler3.call_timestamp

def test_higher_lower_weight(hook):

    lower_priority = create_handler()
    hook.register(lower_priority, priority=-1)

    handler1, handler2, handler3 = handlers = [create_handler() for _ in range(3)] # pylint: disable=redefined-builtin

    higher_priority = create_handler()
    hook.register(higher_priority, priority=1)

    for h in handlers:
        hook.register(h)

    hook.trigger(kwargs={})

    assert handler1.call_timestamp < handler2.call_timestamp < handler3.call_timestamp
    assert lower_priority.call_timestamp > handler3.call_timestamp
    assert higher_priority.call_timestamp < handler1.call_timestamp

def create_handler():
    def returned():
        assert returned.call_timestamp is None
        returned.call_timestamp = next(_call_timestamps)

    returned.call_timestamp = None
    return returned



#pylint: disable=unused-variable
import gossip
import pytest


@pytest.mark.parametrize('should_raise', [True, False])
def test_toggle_hooks(should_raise):

    events = []

    toggle = gossip.Toggle()
    assert toggle.is_off()

    @gossip.register('start')
    def malicious():
        events.append('malicious')
        if should_raise:
            raise CustomException()

    @gossip.register('start', toggles_on=toggle)
    def start():
        events.append('start')

    @gossip.register('end', toggles_off=toggle)
    def end():
        events.append('end')

    if should_raise:
        with pytest.raises(CustomException):
            gossip.trigger('start')
    else:
        gossip.trigger('start')

    assert toggle.is_on() == (not should_raise)

    gossip.trigger('end')

    assert toggle.is_off()

    assert len(events) == len(set(events))
    if should_raise:
        assert 'start' not in events
        assert 'end' not in events
    else:
        assert events.index('malicious') < events.index('start')
        assert 'end' in events


class CustomException(Exception):
    pass
